#!/usr/bin/env python

"""
A really simple virtual processor employing a simple set of instructions which
ignore low-level operations and merely concentrate on variable access, structure
access, structure allocation and function invocations.

Copyright (C) 2007, 2008, 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The execution model of the virtual processor involves the following things:

  * Memory                          contains constants, global variable
                                    references and program code

  * PC (program counter) stack      contains the return address associated
                                    with each function invocation

  * Frame stack                     contains invocation frames in use and in
                                    preparation plus temporary storage

  * Local frame pointer stack       refers to the frames in the frame stack

  * Invocation frame pointer stack

  * Exception handler stack

  * Exception handler locals stack  refers to the state of the local frame
                                    pointer stack

  * Exception handler PC stack      refers to the state of the PC stack

  * Registers: current value,
               boolean status value,
               source value,
               current result,
               current exception,
               current callable
"""

from micropython.program import ReplaceableContext, PlaceholderContext, FragmentObject
import operator

class IllegalInstruction(Exception):
    pass

class IllegalAddress(Exception):
    def __init__(self, address):
        self.address = address
    def __repr__(self):
        return "IllegalAddress(%r)" % self.address
    def __str__(self):
        return repr(self)

class EmptyPCStack(Exception):
    pass

class EmptyFrameStack(Exception):
    pass

class BreakpointReached(Exception):
    pass

class RSVPMachine:

    "A really simple virtual processor."

    def __init__(self, memory, objlist, paramlist, pc=None, debug=0, abort_upon_exception=0):

        """
        Initialise the processor with a 'memory' (a list of values containing
        instructions and data), the object and parameter lists 'objlist' and
        'paramlist', and the optional program counter 'pc'.
        """

        self.memory = memory
        self._objlist = objlist
        self._paramlist = paramlist
        self.objlist = objlist.as_raw()
        self.paramlist = paramlist.as_raw()
        self.library = None

        self.pc = pc or 0
        self.debug = debug
        self.abort_upon_exception = abort_upon_exception

        # Stacks.

        self.pc_stack = []
        self.frame_stack = []
        self.local_sp_stack = [0]
        self.invocation_sp_stack = []
        self.handler_stack = [len(self.memory) - 1] # final handler is the end of the code
        self.handler_local_sp_stack = []
        self.handler_pc_stack = []

        # Registers.

        self.instruction = None
        self.operand = None
        self.value = None
        self.status = None
        self.source = None
        self.callable = None
        self.result = None
        self.exception = None

        # Constants.

        cls = self._get_class("__builtins__", "AttributeError")
        self.attr_error = cls.location
        self.attr_error_instance = cls.instance_template_location
        cls = self._get_class("__builtins__", "TypeError")
        self.type_error = cls.location
        self.type_error_instance = cls.instance_template_location
        cls = self._get_class("__builtins__", "IndexError")
        self.index_error = cls.location
        self.index_error_instance = cls.instance_template_location

        # Debugging attributes.

        self.breakpoints = set()

    def _get_class(self, module, name):
        return self._objlist.access(module, name).get_value()

    # Debugging methods.

    def dump(self):
        print "PC", self.pc, "->", self.load(self.pc)
        print "PC stack", self.pc_stack
        print "Frame stack", self.frame_stack
        print "Local stack pointers", self.local_sp_stack
        print "Invocation stack pointers", self.invocation_sp_stack
        print "Handler stack", self.handler_stack
        print "Handler frame stack", self.handler_local_sp_stack
        print "Handler PC stack", self.handler_pc_stack
        print
        print "Instruction", self.instruction
        print "Operand", self.operand
        print "Value", self.value
        print "Status", self.status
        print "Source", self.source
        print "Callable", self.callable
        print "Result", self.result
        print "Exception", self.exception

    def show(self):
        self.show_memory(self.memory, 0)

    def show_pc(self, run_in=10):
        start = max(0, self.pc - run_in)
        end = self.pc + run_in
        memory = self.memory[start:end]
        self.show_memory(memory, start)

    def show_memory(self, memory, start):
        for i, x in enumerate(memory):
            location = start + i
            if location == self.pc:
                print "->",
            else:
                print "  ",
            print "%5d %r" % (location, x)

    def step(self, dump=0):
        self.execute()
        self.show_pc()
        if dump:
            self.dump()

    def set_break(self, location):
        self.breakpoints.add(location)

    # Internal operations.

    def load(self, address):

        "Return the value at the given 'address'."

        try:
            return self.memory[address]
        except IndexError:
            raise IllegalAddress(address)
        except TypeError:
            raise IllegalAddress(address)

    def save(self, address, value):

        "Save to the given 'address' the specified 'value'."

        try:
            self.memory[address] = value
        except IndexError:
            raise IllegalAddress(address)
        except TypeError:
            raise IllegalAddress(address)

    def new(self, size):

        """
        Allocate space of the given 'size', returning the address of the space.
        """

        addr = len(self.memory)
        for i in range(0, size):
            self.memory.append(None)
        return addr

    def push_pc(self, data):

        "Push 'data' onto the PC stack."

        self.pc_stack.append(data)

    def pull_pc(self):

        "Pull a value from the PC stack and return it."

        try:
            return self.pc_stack.pop()
        except IndexError:
            raise EmptyPCStack

    def run(self):

        "Execute code in the memory, starting from the current PC address."

        breakpoint = 0

        try:
            while 1:
                self.execute()
        except EmptyPCStack:
            pass
        except BreakpointReached:
            breakpoint = 1

        print "Execution terminated",
        if self.exception is not None:
            ref = self.exception
            addr = self.load(ref + 1)
            print "with exception:", self.load(ref)
            print "At address %d: %r" % (addr, self.load(addr))
        elif breakpoint:
            print "with breakpoint."
            print "At address", self.pc
        else:
            print "successfully."

    def test(self, module):

        """
        Test the code in the memory by running the code and investigating the
        contents of variables. Use 'module' to identify result variables.
        """

        self.run()
        success = 1

        if self.exception is None:
            for name in module.keys():
                if name.startswith("result"):
                    label, expected = name.split("_")
                    attr = module[name]

                    # NOTE: Assumptions about headers and content made.

                    attr_location = module.location + 1 + attr.position
                    context, ref = self.load(attr_location)

                    if ref is not None:
                        content = self.load(ref + 1)
                        print label, expected, content
                        success = success and (int(expected) == content)
                    else:
                        print label, expected, "missing"
                        success = 0

            return success
        else:
            return 0

    def execute(self):

        "Execute code in the memory at the current PC address."

        if self.pc in self.breakpoints:
            self.breakpoints.remove(self.pc)
            raise BreakpointReached

        self.instruction = self.load(self.pc)

        # Process any inputs of the instruction.

        self.process_inputs()

        # Perform the instruction itself.

        next_pc = self.perform(self.instruction)

        # Update the program counter.

        if next_pc is None:
            self.pc += 1
        else:
            self.pc = next_pc

    def get_method(self, instruction):

        "Return the handler method for the given 'instruction'."

        instruction_name = instruction.__class__.__name__
        if self.debug:
            print "%8d %s" % (self.pc, instruction_name)
        method = getattr(self, instruction_name, None)
        if method is None:
            raise IllegalInstruction, (self.pc, instruction_name)
        return method

    def perform(self, instruction):

        "Perform the 'instruction', returning the next PC value or None."

        self.operand = instruction.get_operand()
        method = self.get_method(instruction)
        return method()

    def process_inputs(self):

        """
        Process any inputs of the current instruction. This permits any directly
        connected sub-instructions to produce the effects that separate
        instructions would otherwise have.
        """

        value = self.value
        if self.instruction.source is not None:
            self.perform(self.instruction.source)
            self.source = self.value
        self.value = value
        if self.instruction.input is not None:
            self.perform(self.instruction.input)

    def jump(self, addr, next):

        """
        Jump to the subroutine at (or identified by) 'addr'. If 'addr'
        identifies a library function then invoke the library function and set
        PC to 'next' afterwards; otherwise, set PC to 'addr'.
        """

        # Trap library functions introduced through the use of strings instead
        # of proper locations.

        if isinstance(addr, str):
            handler = self.library and self.library.native_functions[addr](self.library)
            if handler is None:
                return next
            else:
                return handler
        else:
            self.push_pc(self.pc + 1)
            return addr

    # Instructions.

    def LoadConst(self):
        self.value = self.operand, self.operand

    def LoadClass(self):
        self.value = PlaceholderContext, self.operand

    def LoadFunction(self):
        self.value = ReplaceableContext, self.operand

    def LoadName(self):
        frame = self.local_sp_stack[-1]
        self.value = self.frame_stack[frame + self.operand]

    def StoreName(self):
        frame = self.local_sp_stack[-1]
        self.frame_stack[frame + self.operand] = self.source # uses the source value

    LoadTemp = LoadName

    def StoreTemp(self):
        frame = self.local_sp_stack[-1]
        self.frame_stack[frame + self.operand] = self.value

    def LoadAddress(self):
        # Preserve context (potentially null).
        self.value = self.load(self.operand)

    def LoadAddressContext(self):
        context, ref = self.load(self.operand)
        inst_context, inst_ref = self.value
        self.value = inst_ref, ref

    def LoadAddressContextCond(self):
        context, ref = self.load(self.operand)
        inst_context, inst_ref = self.value
        self.value = self._LoadAddressContextCond(context, ref, inst_ref)

    def StoreAddress(self):
        # Preserve context.
        self.save(self.operand, self.source)

    def StoreAddressContext(self):
        # Overwrite context if null.
        context_context, context_ref = self.value
        source_context, source_ref = self.source
        if source_context is ReplaceableContext:
            context = context_ref
        else:
            context = source_context
        self.save(self.operand, (context, source_ref))

    def MakeInstance(self):
        size = self.operand
        context, ref = self.value
        # NOTE: Referencing the instance template.
        addr = self._MakeObject(size, ref - 1)
        # Introduce object as context for the new object.
        self.value = addr, addr

    def MakeFragment(self):
        size = self.operand
        addr = self._MakeFragment(size)
        # NOTE: Context is not relevant for fragments.
        self.value = None, addr

    def LoadAttr(self):
        context, ref = self.value
        # Retrieved context should already be appropriate for the instance.
        # NOTE: Adding 1 to skip any header.
        self.value = self.load(ref + self.operand + 1)

    def StoreAttr(self):
        context, ref = self.value
        # Target should already be an instance.
        # NOTE: Adding 1 to skip any header.
        self.save(ref + self.operand + 1, self.source)

    def LoadAttrIndex(self):
        context, ref = self.value
        data = self.load(ref)
        element = self.objlist[data.classcode + self.operand]

        if element is not None:
            attr_index, static_attr, offset = element
            if attr_index == self.operand:
                if static_attr:
                    self.value = self.load(offset) # offset is address of class/module attribute
                else:
                    self.value = self.load(ref + offset)
                return

        self.exception = self._MakeObject(2, self.attr_error_instance)
        return self.RaiseException()

    # LoadAttrIndexContext not defined.

    def LoadAttrIndexContextCond(self):
        inst_context, inst_ref = self.value
        data = self.load(inst_ref)
        element = self.objlist[data.classcode + self.operand]

        if element is not None:
            attr_index, static_attr, offset = element
            if attr_index == self.operand:
                if static_attr:
                    loaded_context, loaded_ref = self.load(offset) # offset is address of class/module attribute
                    if data.attrcode is None: # absent attrcode == class/module
                        self.value = loaded_context, loaded_ref
                    else:
                        self.value = self._LoadAddressContextCond(loaded_context, loaded_ref, inst_ref)
                else:
                    self.value = self.load(inst_ref + offset)
                return

        self.exception = self._MakeObject(2, self.attr_error_instance)
        return self.RaiseException()

    def StoreAttrIndex(self):
        context, ref = self.value
        data = self.load(ref)
        element = self.objlist[data.classcode + self.operand]

        if element is not None:
            attr_index, static_attr, offset = element
            if attr_index == self.operand:
                if static_attr:
                    self.exception = self._MakeObject(2, self.type_error_instance)
                    return self.RaiseException()
                else:
                    self.save(ref + offset, self.source)
                    return

        self.exception = self._MakeObject(2, self.attr_error_instance)
        return self.RaiseException()

    # NOTE: LoadAttrIndexContext is a possibility if a particular attribute can always be overridden.

    def MakeFrame(self):
        self.invocation_sp_stack.append(len(self.frame_stack))
        self.frame_stack.extend([None] * self.operand)

    def DropFrame(self):
        self.local_sp_stack.pop()
        frame = self.invocation_sp_stack.pop()
        del self.frame_stack[frame:] # reset stack before call

    def StoreFrame(self):
        frame = self.invocation_sp_stack[-1] # different from the current frame after MakeFrame
        self.frame_stack[frame + self.operand] = self.value

    def StoreFrameIndex(self):
        context, ref = self.value
        frame = self.invocation_sp_stack[-1] # different from the current frame after MakeFrame
        data = self.load(ref)
        element = self.paramlist[data.funccode + self.operand]

        if element is not None:
            # NOTE: Need to ensure correct positioning where a context has been generated.
            param_index, offset = element
            if param_index == self.operand:
                self.frame_stack[frame + offset] = self.source
                return

        self.exception = self._MakeObject(2, self.type_error_instance)
        return self.RaiseException()

    def LoadCallable(self):
        context, ref = self.value
        data = self.load(ref)
        self.callable = data.codeaddr

    def StoreCallable(self):
        context, ref = self.value
        # NOTE: Should improve the representation and permit direct saving.
        data = self.load(ref)
        self.save(ref, data.with_callable(self.callable))

    def LoadContext(self):
        context, ref = self.value
        # NOTE: Omission of the context of the context would make things like
        # NOTE: self() inside methods impossible.
        self.value = context, context

    def CheckContext(self):
        self.status = self.value[1] is not ReplaceableContext

    def CheckClass(self):
        context, ref = self.value
        if ref in (ReplaceableContext, PlaceholderContext):
            self.status = 0
            return

        data = self.load(ref)

        # Classes are not themselves usable as the self argument.
        # NOTE: This may change at some point.
        # However, where classes appear as the context, instance
        # compatibility is required in the first argument.

        self.status = data.attrcode is None # absent attrcode == class

    def CheckFrame(self):
        (nargs, ndefaults, has_star) = self.operand

        # The frame is actually installed as the locals.
        # Retrieve the context from the first local.

        frame = self.local_sp_stack[-1]
        nlocals = len(self.frame_stack[frame:])

        # NOTE: Not testing (nlocals <= nargs or has_star) due to imprecise
        # NOTE: invocation frame removal (after frame adjustment).

        if not ((nargs - ndefaults) <= nlocals):
            raise Exception, "CheckFrame %r (%r <= %r <= %r)" % (self.operand, nargs - ndefaults, nlocals, nargs)
            self.exception = self._MakeObject(2, self.type_error_instance)
            return self.RaiseException()

    def FillDefaults(self):
        context, ref = self.value
        (nargs, ndefaults) = self.operand

        # The frame is actually installed as the locals.

        frame = self.local_sp_stack[-1]
        nlocals = len(self.frame_stack[frame:])

        # Support population of defaults.
        # This involves copying the "attributes" of a function into the frame.

        default = nlocals - (nargs - ndefaults)
        self.frame_stack.extend([None] * (nargs - nlocals))
        pos = nlocals

        while pos < nargs:
            self.frame_stack[frame + pos] = self.load(ref + default + 1) # skip header
            default += 1
            pos += 1

    def CheckSelf(self):
        context, ref = self.value
        context_context, context_ref = self.source

        # Check the details of the proposed context and the target's context.

        self.status = self._CheckInstance(ref, context_ref)

    def JumpInFrame(self):
        codeaddr = self.callable
        return self.jump(codeaddr, self.pc + 1) # return to the instruction after this one

    def JumpWithFrame(self):
        codeaddr = self.callable
        self.local_sp_stack.append(self.invocation_sp_stack[-1]) # adopt the invocation frame
        return self.jump(codeaddr, self.pc + 1) # return to the instruction after this one

    def JumpWithFrameDirect(self):
        operand = self.operand
        self.local_sp_stack.append(self.invocation_sp_stack[-1]) # adopt the invocation frame
        return self.jump(operand, self.pc + 1) # return to the instruction after this one

    def ExtendFrame(self):
        self.frame_stack.extend([None] * self.operand)

    def AdjustFrame(self):
        self.invocation_sp_stack[-1] += self.operand

    def Return(self):
        return self.pull_pc()

    def LoadResult(self):
        self.value = self.result

    def StoreResult(self):
        self.result = self.value

    def Jump(self):
        return self.operand

    def JumpIfTrue(self):
        if self.status:
            return self.operand

    def JumpIfFalse(self):
        if not self.status:
            return self.operand

    def LoadException(self):
        self.value = self.exception, self.exception

    def StoreException(self):
        self.exception = self.value[1]

    def ClearException(self):
        self.exception = None

    def RaiseException(self):
        # NOTE: Adding the program counter as the first attribute.
        self.save(self.exception + 1, self.pc)
        # Jumping to the current handler.
        if self.abort_upon_exception:
            raise Exception
        return self.handler_stack[-1]

    def PushHandler(self):
        self.handler_stack.append(self.operand)
        self.handler_local_sp_stack.append(len(self.local_sp_stack))
        self.handler_pc_stack.append(len(self.pc_stack))

    def PopHandler(self):
        # Reduce the local frame pointer stack to refer to the handler's frame.
        del self.local_sp_stack[self.handler_local_sp_stack.pop():]
        # Reduce the PC stack to discard all superfluous return addresses.
        self.pc_stack = self.pc_stack[:self.handler_pc_stack.pop()]
        self.handler_stack.pop()

    def CheckException(self):
        self.status = self.exception is not None and self._CheckInstance(self.exception, self.value[1])

    def TestIdentity(self):
        self.status = self.value[1] == self.source[1]

    def TestIdentityAddress(self):
        self.status = self.value[1] == self.operand

    # LoadBoolean is implemented in the generated code.
    # StoreBoolean is implemented by testing against the True value.

    def InvertBoolean(self):
        self.status = not self.status

    # Common implementation details.

    def _CheckInstance(self, ref, cls):
        data = self.load(ref)
        target_data = self.load(cls)

        # Insist on instance vs. class.

        if data.attrcode is None: # absent attrcode == class/module
            return 0

        if target_data.attrcode is not None: # present attrcode == instance
            return 0

        # Find the table entry for the descendant.

        element = self.objlist[target_data.classcode + data.attrcode]

        if element is not None:
            attr_index, static_attr, offset = element
            return attr_index == data.attrcode
        else:
            return 0

    def _MakeObject(self, size, ref):
        # Load the template.
        data = self.load(ref)
        addr = self.new(size)
        # Save the header, overriding the size.
        self.save(addr, data.with_size(size))
        return addr

    def _MakeFragment(self, size):
        # Reserve twice the amount of space.
        addr = self.new(size * 2)
        # Save the header, overriding the size.
        self.save(addr, FragmentObject(size, size * 2))
        return addr

    def _LoadAddressContextCond(self, context, ref, inst_ref):
        # Check the instance context against the target's context.
        # This provides the context overriding for methods.
        if context is ReplaceableContext or context is not PlaceholderContext and self._CheckInstance(inst_ref, context):
            # Replace the context with the instance.
            return inst_ref, ref
        else:
            return context, ref

class Library:

    "Native function implementations."

    def __init__(self, machine, true_constant, false_constant):

        """
        Initialise the library with the 'machine' and the addresses
        'true_constant' and 'false_constant'.
        """

        self.machine = machine
        self.true_constant = true_constant
        self.false_constant = false_constant

        # Native class constants.

        cls = self.machine._get_class("__builtins__", "int")
        self.int_class = cls.location
        self.int_instance = cls.instance_template_location
        cls = self.machine._get_class("__builtins__", "list")
        self.list_class = cls.location
        self.list_instance = cls.instance_template_location
        cls = self.machine._get_class("__builtins__", "tuple")
        self.tuple_class = cls.location

        self.type_error_instance = self.machine.type_error_instance
        self.frame_stack = self.machine.frame_stack
        self.local_sp_stack = self.machine.local_sp_stack

    def builtins_int_arithmetic_op(self, op):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_context, left = self.frame_stack[frame]
        right_context, right = self.frame_stack[frame + 1]

        # Test operand suitability.
        # NOTE: Support other types.

        if not (self.machine._CheckInstance(left, self.int_class) and self.machine._CheckInstance(right, self.int_class)):
            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for data.

        left_data = left + 1
        right_data = right + 1

        # Make a new object.

        addr = self.machine._MakeObject(2, self.int_instance)

        # Store the result.
        # NOTE: The data is considered ready to use.

        self.machine.save(addr + 1, op(self.machine.load(left_data), self.machine.load(right_data)))

        # Return the new object.
        # Introduce object as context for the new object.

        self.machine.result = addr, addr

    def builtins_int_add(self):
        return self.builtins_int_arithmetic_op(operator.add)

    def builtins_int_sub(self):
        return self.builtins_int_arithmetic_op(operator.sub)

    def builtins_int_bool(self):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_context, left = self.frame_stack[frame]

        # Test operand suitability.

        if not self.machine._CheckInstance(left, self.int_class):
            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for data.

        left_data = left + 1

        # Test the data.
        # NOTE: The data is considered ready to use.

        if self.machine.load(left_data) != 0:
            self.machine.result = self.true_constant, self.true_constant
        else:
            self.machine.result = self.false_constant, self.false_constant

    def builtins_int_neg(self):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_context, left = self.frame_stack[frame]

        # Test operand suitability.

        if not self.machine._CheckInstance(left, self.int_class):
            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for data.

        left_data = left + 1

        # Make a new object.

        addr = self.machine._MakeObject(2, self.int_instance)

        # Store the result.
        # NOTE: The data is considered ready to use.

        self.machine.save(addr + 1, -self.machine.load(left_data))

        # Return the new object.
        # Introduce object as context for the new object.

        self.machine.result = addr, addr

    def builtins_int_op(self, op, true_if_incompatible):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_context, left = self.frame_stack[frame]
        right_context, right = self.frame_stack[frame + 1]

        # Test operand suitability.
        # NOTE: Support other types.
        # NOTE: Handle comparisons of incompatible types more appropriately.

        if not (self.machine._CheckInstance(left, self.int_class) and self.machine._CheckInstance(right, self.int_class)):
            if true_if_incompatible:
                self.machine.result = self.true_constant, self.true_constant
            else:
                self.machine.result = self.false_constant, self.false_constant
            return

        # NOTE: Assume single location for data.

        left_data = left + 1
        right_data = right + 1

        # Test the data.
        # NOTE: The data is considered ready to use.

        if op(self.machine.load(left_data), self.machine.load(right_data)):
            self.machine.result = self.true_constant, self.true_constant
        else:
            self.machine.result = self.false_constant, self.false_constant

    def builtins_int_lt(self):
        return self.builtins_int_op(operator.lt, 0)

    def builtins_int_le(self):
        return self.builtins_int_op(operator.le, 0)

    def builtins_int_gt(self):
        return self.builtins_int_op(operator.gt, 0)

    def builtins_int_ge(self):
        return self.builtins_int_op(operator.ge, 0)

    def builtins_int_eq(self):
        return self.builtins_int_op(operator.eq, 0)

    def builtins_int_ne(self):
        return self.builtins_int_op(operator.ne, 1)

    def builtins_bool_bool(self):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_context, left = self.frame_stack[frame]
        self.machine.result = left, left

    def builtins_list_new(self):
        frame = self.local_sp_stack[-1]

        # The first parameter should be empty.
        # NOTE: Specific copying of tuples/lists.

        args_context, args = self.frame_stack[frame + 1]

        # Test operand suitability.

        if self.machine._CheckInstance(args, self.list_class):
            _x, sequence = self.machine.load(args + 1)
            header = self.machine.load(sequence)
            size = header.occupied_size
        elif self.machine._CheckInstance(args, self.tuple_class):
            sequence = args
            header = self.machine.load(sequence)
            size = header.size
        else:
            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # Copy the sequence contents.

        new_fragment = self.machine._MakeFragment(size)
        for i in range(1, size):
            self.machine.save(new_fragment + i, self.machine.load(sequence + i))

        # Make the list instance.

        addr = self.machine._MakeObject(2, self.list_instance)
        self.machine.save(addr + 1, (None, new_fragment))

        self.machine.result = addr, addr

    def builtins_list_getitem(self):
        frame = self.local_sp_stack[-1]

        # Get the operand address.

        item_context, item = self.frame_stack[frame + 1]

        # Get the list address.

        obj_context, obj = self.frame_stack[frame]

        # Get the fragment address.
        # NOTE: Assume single location for header.

        _x, fragment = self.machine.load(obj + 1)

        # Get the fragment header.

        header = self.machine.load(fragment)
        nelements = header.occupied_size - 1

        # NOTE: Assume single location for data and header.

        item_pos = self.machine.load(item + 1)
        if item_pos >= 0 and item_pos < nelements:
            pass
        elif item_pos < 0 and item_pos >= -nelements:
            item_pos = nelements + item_pos
        else:
            self.machine.exception = self.machine._MakeObject(2, self.index_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for header.

        self.machine.result = self.machine.load(fragment + 1 + item_pos)

    def builtins_list_len(self):
        frame = self.local_sp_stack[-1]

        # Get the list address.

        obj_context, obj = self.frame_stack[frame]

        # Get the fragment address.
        # NOTE: Assume single location for header.

        _x, fragment = self.machine.load(obj + 1)

        # Get the fragment header.

        header = self.machine.load(fragment)
        nelements = header.occupied_size - 1

        # Make a new object.

        addr = self.machine._MakeObject(2, self.int_instance)

        # Store the result.
        # NOTE: The data is considered ready to use.

        self.machine.save(addr + 1, nelements)

        # Return the new object.
        # Introduce object as context for the new object.

        self.machine.result = addr, addr

    def builtins_list_append(self):
        frame = self.local_sp_stack[-1]

        # Get operand address.

        arg_context, arg = self.frame_stack[frame + 1]

        # Get the list address.

        obj_context, obj = self.frame_stack[frame]

        # Get the fragment address.
        # NOTE: Assume single location for header.

        _x, fragment = self.machine.load(obj + 1)

        # Get the fragment header.

        header = self.machine.load(fragment)

        # Attempt to add the reference.

        if header.occupied_size < header.allocated_size:
            self.machine.save(fragment + header.occupied_size, (arg_context, arg))
            header.occupied_size += 1
        else:

            # Make a new fragment, maintaining more space than currently
            # occupied in order to avoid reallocation.

            new_fragment = self.machine._MakeFragment(header.allocated_size)

            # Copy existing elements.

            for i in range(1, header.allocated_size):
                self.machine.save(new_fragment + i, self.machine.load(fragment + i))

            self.machine.save(new_fragment + header.allocated_size, (arg_context, arg))

            # Set the new fragment in the object.
            # NOTE: The old fragment could be deallocated.

            self.machine.save(obj + 1, (None, new_fragment))

    def builtins_object_init(self):
        pass

    native_functions = {

        # Native method implementations:

        "__builtins__.int.__add__" : builtins_int_add,
        "__builtins__.int.__radd__" : builtins_int_add,               # NOTE: To be made distinct.
        "__builtins__.int.__sub__" : builtins_int_sub,
        "__builtins__.int.__iadd__" : builtins_int_add,
        "__builtins__.int.__bool__" : builtins_int_bool,
        "__builtins__.int.__neg__" : builtins_int_neg,
        "__builtins__.int.__lt__" : builtins_int_lt,
        "__builtins__.int.__le__" : builtins_int_le,
        "__builtins__.int.__gt__" : builtins_int_gt,
        "__builtins__.int.__ge__" : builtins_int_ge,
        "__builtins__.int.__eq__" : builtins_int_eq,
        "__builtins__.int.__ne__" : builtins_int_ne,
        "__builtins__.bool.__bool__" : builtins_bool_bool,
        "__builtins__.list.__getitem__" : builtins_list_getitem,
        "__builtins__.list.__len__" : builtins_list_len,
        "__builtins__.list.append" : builtins_list_append,
        "__builtins__.object.__init__" : builtins_object_init,        # NOTE: A no-operation.
        "__builtins__.BaseException.__init__" : builtins_object_init, # NOTE: To be made distinct, potentially in the builtins module.

        # Native instantiators:

        "__builtins__.list" : builtins_list_new,
        }

# Convenience functions.

def machine(program, with_builtins=0, debug=0, abort_upon_exception=0):
    print "Making the image..."
    code = program.get_image(with_builtins)
    print "Getting raw structures..."
    ot = program.get_object_table()
    pt = program.get_parameter_table()
    objlist = ot.as_list()
    paramlist = pt.as_list()
    print "Getting raw image..."
    rc = program.get_raw_image()
    print "Initialising the machine..."
    importer = program.get_importer()
    true_constant = importer.get_constant(True).location
    false_constant = importer.get_constant(False).location
    rm = RSVPMachine(rc, objlist, paramlist, debug=debug, abort_upon_exception=abort_upon_exception)
    library = Library(rm, true_constant, false_constant)
    rm.library = library
    rm.pc = program.code_location
    print "Returning program occupying %d locations." % len(rm.memory)
    return rm

# vim: tabstop=4 expandtab shiftwidth=4
